from __future__ import annotations

import copy
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Final, Literal, overload

from loguru import logger

from ..assets.hsr.file_paths import SOURCE_TO_PATH
from ..assets.hsr.manager import AssetManager
from ..assets.updater import AssetUpdater
from ..constants.hsr import DEFAULT_STATS
from ..enums.enum import Game
from ..enums.hsr import Element, Language, Path, StatType, TraceType
from ..models.hsr import CharacterIcon, LightConeIcon, Player, ShowcaseResponse, Stat
from ..models.hsr.build import Build
from ..models.hsr.character import Eidolon
from ..utils import update_stats
from .base import BaseClient

if TYPE_CHECKING:
    from ..models.enka.owner import Owner
    from ..models.hsr.character import Character, LightCone, Relic, Trace
    from .cache import BaseTTLCache

__all__ = ("HSRClient",)

API_URL: Final[str] = "https://enka.network/api/hsr/uid/{uid}"


class HSRClient(BaseClient):
    """The main client to interact with the Enka Network Honkai Star Rail API.

    Args:
        lang (Language | str): The language to use for the client, defaults to Language.ENGLISH.
        headers (dict[str, Any] | None): The headers to use for the client, defaults to None.
        cache (BaseTTLCache | None): The cache to use for the client, defaults to None.
        use_enka_icons (bool): Whether to get stat icons from Enka, defaults to True.
    """

    def __init__(
        self,
        lang: Language | str = Language.ENGLISH,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
        use_enka_icons: bool = True,
    ) -> None:
        super().__init__(Game.HSR, headers=headers, cache=cache)

        if isinstance(lang, str):
            try:
                lang = Language(lang)
            except ValueError as e:
                available_langs = ", ".join(lang.value for lang in Language)
                msg = f"Invalid language: {lang}, must be one of {available_langs}"
                raise ValueError(msg) from e

        self._lang = lang
        self._use_enka_icons = use_enka_icons

    async def __aenter__(self) -> HSRClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    def _check_assets(self) -> None:
        if self._assets is None:
            msg = f"Client is not started, call `{self.__class__.__name__}.start` first"
            raise RuntimeError(msg)

    def _post_process_relic(self, relic: Relic) -> None:
        text_map = self._assets.text_map
        relic.set_name = text_map[relic.set_name]

        relic_data = self._assets.relic_data[str(relic.id)]
        relic.icon = self._get_icon(relic_data["Icon"])
        relic.rarity = relic_data["Rarity"]

        for stat in relic.stats:
            stat.name = text_map[stat.type.value]
            stat.icon = self._get_icon(
                self._assets.property_config_data[stat.type.value], enka=self._use_enka_icons
            )

    def _post_process_light_cone(self, light_cone: LightCone) -> None:
        text_map = self._assets.text_map
        light_cone.name = text_map[light_cone.name]
        light_cone.rarity = self._assets.light_cones_data[str(light_cone.id)]["Rarity"]
        light_cone.icon = LightConeIcon(light_cone_id=light_cone.id)

        for stat in light_cone.stats:
            stat.name = text_map[stat.type.value]
            stat.icon = self._get_icon(
                self._assets.property_config_data[stat.type.value], enka=self._use_enka_icons
            )

    def _post_process_trace(self, trace: Trace, unlocked_eidolon_ids: list[int]) -> None:
        skill_tree_data = self._assets.skill_tree_data
        trace_data = skill_tree_data[str(trace.id)]

        try:
            skill_ids: list[int] = trace_data["skillIds"]
        except KeyError as e:
            msg = "Skill IDs not found in trace data, please update the assets with `update_assets`"
            raise RuntimeError(msg) from e

        trace.anchor = trace_data["anchor"]
        trace.icon = self._get_icon(
            trace_data["icon"], enka=self._use_enka_icons or "SkillIcons" in trace_data["icon"]
        )
        trace.type = TraceType(trace_data["pointType"])
        trace.max_level = trace_data["maxLevel"]

        for eidolon_id in unlocked_eidolon_ids:
            eidolon_data = self._assets.eidolon_data[str(eidolon_id)]

            for skill_id in skill_ids:
                if str(skill_id) in eidolon_data["SkillAddLevelList"]:
                    trace.level += eidolon_data["SkillAddLevelList"][str(skill_id)]
                    trace.boosted = True
                    break

    def _post_process_character(self, character: Character) -> None:
        self._check_assets()

        character.icon = CharacterIcon(character_id=character.id)

        character_data = self._assets.character_data[str(character.id)]
        text_map_hash: int = character_data["AvatarName"]["Hash"]
        character.name = self._assets.text_map[str(text_map_hash)]
        character.rarity = character_data["Rarity"]
        character.element = Element(character_data["Element"])
        character.path = Path(character_data["AvatarBaseType"])

        # Eidolons
        eidolon_ids: list[int] = character_data["RankIDList"]
        for i, eidolon_id in enumerate(eidolon_ids, start=1):
            eidolon_data = self._assets.eidolon_data[str(eidolon_id)]
            character.eidolons.append(
                Eidolon(
                    id=eidolon_id,
                    icon=self._get_icon(eidolon_data["IconPath"]),
                    unlocked=i <= character.eidolons_unlocked,
                )
            )

        for trace in character.traces:
            self._post_process_trace(
                trace, [e.id for e in character.eidolons[: character.eidolons_unlocked]]
            )
        for relic in character.relics:
            self._post_process_relic(relic)
        if character.light_cone is not None:
            self._post_process_light_cone(character.light_cone)

        # Credits to Algoinde for the following code
        chara_stats = self._add_up_character_stats(character)
        final_stats: dict[StatType, float] = {
            StatType.MAX_HP: (
                chara_stats["BaseHP"]
                + chara_stats["HPBase"]
                + chara_stats["HPAdd"] * (character.level - 1)
            )
            * (1 + chara_stats["HPAddedRatio"])
            + chara_stats["HPDelta"]
            + chara_stats["HPConvert"],
            StatType.ATK: (
                chara_stats["BaseAttack"]
                + chara_stats["AttackBase"]
                + chara_stats["AttackAdd"] * (character.level - 1)
            )
            * (1 + chara_stats["AttackAddedRatio"])
            + chara_stats["AttackDelta"]
            + chara_stats["AttackConvert"],
            StatType.DEF: (
                chara_stats["BaseDefence"]
                + chara_stats["DefenceBase"]
                + chara_stats["DefenceAdd"] * (character.level - 1)
            )
            * (1 + chara_stats["DefenceAddedRatio"])
            + chara_stats["DefenceDelta"]
            + chara_stats["DefenceConvert"],
            StatType.SPD: (chara_stats["BaseSpeed"] + chara_stats["SpeedBase"])
            * (1 + chara_stats["SpeedAddedRatio"])
            + chara_stats["SpeedDelta"]
            + chara_stats["SpeedConvert"],
            StatType.BASE_HP: chara_stats["BaseHP"]
            + chara_stats["HPBase"]
            + chara_stats["HPAdd"] * (character.level - 1),
            StatType.BASE_ATK: chara_stats["BaseAttack"]
            + chara_stats["AttackBase"]
            + chara_stats["AttackAdd"] * (character.level - 1),
            StatType.BASE_DEF: chara_stats["BaseDefence"]
            + chara_stats["DefenceBase"]
            + chara_stats["DefenceAdd"] * (character.level - 1),
            StatType.BASE_SPEED: chara_stats["BaseSpeed"] + chara_stats["SpeedBase"],
            StatType.CRIT_RATE: chara_stats["CriticalChanceBase"] + chara_stats["CriticalChance"],
            StatType.CRIT_DMG: chara_stats["CriticalDamageBase"] + chara_stats["CriticalDamage"],
            StatType.BREAK_EFFECT: chara_stats["BreakDamageAddedRatioBase"]
            + chara_stats["BreakDamageAddedRatio"],
            StatType.HEALING_BOOST: chara_stats["HealRatioBase"] + chara_stats["HealRatioConvert"],
            StatType.ENERGY_REGEN_RATE: chara_stats["SPRatioBase"]
            + chara_stats["SPRatio"]
            + chara_stats["SPRatioConvert"]
            + 1,
            StatType.EFFECT_HIT_RATE: chara_stats["StatusProbabilityBase"]
            + chara_stats["StatusProbability"]
            + chara_stats["StatusProbabilityConvert"],
            StatType.EFFECT_RES: chara_stats["StatusResistanceBase"]
            + chara_stats["StatusResistance"]
            + chara_stats["StatusResistanceConvert"],
            StatType.PHYSICAL_DMG_BOOST: chara_stats["PhysicalAddedRatio"],
            StatType.FIRE_DMG_BOOST: chara_stats["FireAddedRatio"],
            StatType.ICE_DMG_BOOST: chara_stats["IceAddedRatio"],
            StatType.LIGHTNING_DMG_BOOST: chara_stats["ThunderAddedRatio"],
            StatType.WIND_DMG_BOOST: chara_stats["WindAddedRatio"],
            StatType.QUANTUM_DMG_BOOST: chara_stats["QuantumAddedRatio"],
            StatType.IMAGINARY_DMG_BOOST: chara_stats["ImaginaryAddedRatio"],
        }

        character.stats = {
            stat_type: Stat(
                type=stat_type,
                value=value,
                name=self._assets.text_map[stat_type.value],
                icon=self._get_icon(
                    self._assets.property_config_data[stat_type.value], enka=self._use_enka_icons
                ),
            )
            for stat_type, value in final_stats.items()
        }

    def _add_up_character_stats(self, character: Character) -> dict[str, float]:  # noqa: C901
        chara_stats = DEFAULT_STATS.copy()

        # Add base stats
        chara_meta: dict[str, float] = self._assets.meta_data["avatar"][str(character.id)][
            str(character.ascension)
        ]
        update_stats(chara_stats, chara_meta)

        # Add light cone stats
        if character.light_cone is not None:
            lc = character.light_cone

            lc_meta: dict[str, float] = self._assets.meta_data["equipment"][str(lc.id)][
                str(lc.ascension)
            ]
            update_stats(chara_stats, lc_meta)

            # Add light cone skill stats
            lc_skill_meta_ = self._assets.meta_data["equipmentSkill"]
            if str(lc.id) in lc_skill_meta_:
                lc_skill_meta: dict[str, float] = lc_skill_meta_[str(lc.id)][str(lc.superimpose)][
                    "props"
                ]
                update_stats(chara_stats, lc_skill_meta)

        # Add relic stats
        relic_sets: defaultdict[int, int] = defaultdict(int)
        for relic in character.relics:
            relic_sets[relic.set_id] += 1

            for stat in relic.stats:
                update_stats(chara_stats, {stat.type.value: stat.value})

        # Add relic set stats
        for relic_set_id, count in relic_sets.items():
            for i in range(1, count + 1):
                relic_set_meta: dict[str, dict[str, float]] = self._assets.meta_data["relic"][
                    "setSkill"
                ][str(relic_set_id)].get(str(i))
                if relic_set_meta is not None:
                    update_stats(chara_stats, relic_set_meta["props"])

        # Add trace stats
        for trace in character.traces:
            trace_meta_ = self._assets.meta_data["tree"]
            if str(trace.id) in trace_meta_:
                trace_meta: dict[str, float] = trace_meta_[str(trace.id)][str(trace.level)]["props"]
                update_stats(chara_stats, trace_meta)

        return chara_stats

    def _post_process_player(self, player: Player) -> None:
        self._check_assets()

        player.icon = self._get_icon(self._assets.avatar_data[player.icon]["Icon"])

    def _post_process_showcase(self, showcase: ShowcaseResponse) -> None:
        for character in showcase.characters:
            self._post_process_character(character)
        self._post_process_player(showcase.player)

    def _get_icon(self, icon: str, *, enka: bool = True) -> str:
        icon = icon.replace(".png", "")
        if enka:
            return f"https://enka.network/ui/hsr/{icon}.png"
        return f"https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/icon/property/{icon.split('/')[-1]}.png"

    async def start(self) -> None:
        """Start the client."""
        await super().start()
        assert self._session is not None

        self._assets = AssetManager(self._lang)
        self._asset_updater = AssetUpdater(self._session, SOURCE_TO_PATH, self._lang)

        loaded = await self._assets.load()
        if not loaded:
            await self.update_assets()

    async def update_assets(self) -> None:
        """Update game assets."""
        if self._asset_updater is None or self._assets is None:
            msg = f"Client is not started, call `{self.__class__.__name__}.start` first"
            raise RuntimeError(msg)

        logger.info("Updating HSR assets...")
        await self._asset_updater.update()
        await self._assets.load()
        logger.info("HSR assets updated")

    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[False] = False
    ) -> ShowcaseResponse: ...
    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[True] = True
    ) -> dict[str, Any]: ...
    async def fetch_showcase(
        self, uid: str | int, *, raw: bool = False
    ) -> ShowcaseResponse | dict[str, Any]:
        """Fetches the Impact character showcase of the given UID.

        Args:
            uid (str | int): The UID of the user.
            raw (bool): Whether to return the raw data or not, defaults to False.

        Returns:
            ShowcaseResponse | dict[str, Any]: The parsed or raw showcase data.
        """
        url = API_URL.format(uid=uid)

        data = await self._request(url)
        if raw:
            return data

        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    def parse_showcase(self, data: dict[str, Any]) -> ShowcaseResponse:
        """Parses the given showcase data.

        Args:
            data (dict[str, Any]): The showcase data.

        Returns:
            ShowcaseResponse: The parsed showcase data.
        """
        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    async def fetch_builds(self, owner: Owner) -> dict[str, list[Build]]:
        """Fetches the builds of the given owner.

        Args:
            owner (Owner): The owner of the builds.

        Returns:
            dict[str, list[Build]]: Character ID to list of builds mapping.
        """
        url = f"https://enka.network/api/profile/{owner.username}/hoyos/{owner.hash}/builds/"
        data = await self._request(url)
        result: dict[str, list[Build]] = {}

        for key, builds in data.items():
            result[key] = []
            for build in builds:
                build_ = Build(**build)
                self._post_process_character(build_.character)
                result[key].append(build_)

        return result



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import aiofiles
import orjson
from loguru import logger

from ..errors import AssetUpdateError

if TYPE_CHECKING:
    import aiohttp

    from ..enums import gi, hsr

__all__ = ("AssetUpdater",)


class AssetUpdater:
    """Game asset updater."""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        source_to_path: dict[str, str],
        lang: gi.Language | hsr.Language,
    ) -> None:
        self._session = session
        self._source_to_path = source_to_path
        self._lang = lang

    async def _fetch_json(self, url: str) -> Any:
        logger.debug(f"Fetching {url}")

        async with self._session.get(url) as resp:
            if resp.status != 200:
                raise AssetUpdateError(resp.status, url)

            return orjson.loads(await resp.read())

    async def update(self) -> None:
        """Update all assets."""
        for source, path in self._source_to_path.items():
            path_ = path.format(lang=self._lang.value)
            if not os.path.exists(os.path.dirname(path_)):
                os.makedirs(os.path.dirname(path_))

            data = await self._fetch_json(source.format(lang=self._lang.value))

            async with aiofiles.open(path_, "w", encoding="utf-8") as f:
                bytes_ = orjson.dumps(data)
                await f.write(bytes_.decode())



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator

__all__ = ("Player", "PlayerStats")


class PlayerStats(BaseModel):
    """Represents the statistics of a HSR player.

    Attributes:
        achievement_count (int): The number of achievements the player has.
        light_cone_count (int): The number of light cones the player has.
        character_count (int): The number of characters the player has.
        max_simulated_universe_world (int, optional): The maximum world completed in the simulated universe.
        book_count (int, optional): The number of books the player has. (Added in game version 2.2)
        relic_count (int, optional): The number of relics the player has. (Added in game version 2.2)
        music_count (int, optional): The number of music tracks the player has. (Added in game version 2.2)
    """

    achievement_count: int = Field(0, alias="achievementCount")
    light_cone_count: int = Field(0, alias="equipmentCount")
    character_count: int = Field(alias="avatarCount")
    max_simulated_universe_world: int | None = Field(None, alias="maxRogueChallengeScore")

    # New after game version 2.2
    book_count: int | None = Field(None, alias="bookCount")
    relic_count: int | None = Field(None, alias="relicCount")
    music_count: int | None = Field(None, alias="musicCount")


class Player(BaseModel):
    """Represents a HSR player.

    Attributes:
        nickname (str): The player's nickname.
        signature (str): The player's signature.
        uid (int): The player's UID.
        level (int): The player's level.
        equilibrium_level (int): The player's equilibrium level.
        friend_count (int): The number of friends the player has.
        stats (:class:`PlayerStats`): The player's statistics.
    """

    nickname: str
    signature: str = ""
    uid: int
    level: int
    equilibrium_level: int = Field(0, alias="worldLevel")
    friend_count: int = Field(0, alias="friendCount")
    stats: PlayerStats = Field(alias="recordInfo")

    # The following fields are added in post-processing
    icon: str = Field("", alias="headIcon")

    @field_validator("icon", mode="before")
    def _stringify_icon(cls, value: int) -> str:
        return str(value)



from __future__ import annotations

from pydantic import BaseModel, Field

from .character import Character


class Build(BaseModel):
    """Represents a Genshin Impact build.

    Attributes:
        id (int): The build's ID.
        name (str): The build's name.
        order (int): The build's order.
        live (bool): Whether the build is live.
        character_id (int): The build's character ID.
        character (:class:`Character`): The build's character data.
    """

    id: int
    name: str
    order: int
    live: bool
    character_id: int = Field(alias="avatar_id")
    character: Character = Field(alias="avatar_data")



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data import AssetData
from .file_paths import (
    AVATAR_DATA_PATH,
    CHARACTER_DATA_PATH,
    EIDOLON_DATA_PATH,
    LIGHT_CONE_DATA_PATH,
    META_DATA_PATH,
    PROPERTY_CONFIG_PATH,
    RELIC_DATA_PATH,
    SKILL_TREE_DATA_PATH,
    TEXT_MAP_PATH,
)

if TYPE_CHECKING:
    from ...enums.hsr import Language

__all__ = ("AssetManager",)


class AssetManager:
    """Honkai Star Rail asset manager."""

    def __init__(self, lang: Language) -> None:
        self._lang = lang

        self.text_map = TextMap(lang)
        self.character_data = CharacterData()
        self.skill_tree_data = SkillTreeData()
        self.light_cones_data = LightConesData()
        self.relic_data = RelicData()
        self.meta_data = MetaData()
        self.avatar_data = AvatarData()
        self.property_config_data = PropertyConfigData()
        self.eidolon_data = EidolonData()

    async def load(self) -> bool:
        """Load all assets.

        Returns:
            bool: Whether all assets were loaded successfully.
        """
        return (
            await self.text_map.load()
            and await self.character_data.load()
            and await self.skill_tree_data.load()
            and await self.light_cones_data.load()
            and await self.relic_data.load()
            and await self.meta_data.load()
            and await self.avatar_data.load()
            and await self.property_config_data.load()
            and await self.eidolon_data.load()
        )


class TextMap(AssetData):
    def __init__(self, lang: Language) -> None:
        super().__init__()
        self._lang = lang

    async def load(self) -> bool:
        text_map = await self._open_json(TEXT_MAP_PATH)
        if text_map is not None:
            self._data = text_map[self._lang.value]
        return self._data is not None


class CharacterData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(CHARACTER_DATA_PATH)
        return self._data is not None


class SkillTreeData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(SKILL_TREE_DATA_PATH)
        return self._data is not None


class LightConesData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(LIGHT_CONE_DATA_PATH)
        return self._data is not None


class RelicData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(RELIC_DATA_PATH)
        return self._data is not None


class MetaData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(META_DATA_PATH)
        return self._data is not None


class AvatarData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(AVATAR_DATA_PATH)
        return self._data is not None


class PropertyConfigData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(PROPERTY_CONFIG_PATH)
        return self._data is not None


class EidolonData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(EIDOLON_DATA_PATH)
        return self._data is not None



from __future__ import annotations

import abc
import os
import pathlib
import time

import aiosqlite

__all__ = ("BaseTTLCache", "MemoryCache", "SQLiteCache")


class BaseTTLCache(abc.ABC):
    @abc.abstractmethod
    async def start(self) -> None: ...

    @abc.abstractmethod
    async def close(self) -> None: ...

    @abc.abstractmethod
    async def get(self, key: str) -> str | None: ...

    @abc.abstractmethod
    async def set(self, key: str, value: str, ttl: int) -> None: ...

    @abc.abstractmethod
    async def delete(self, key: str) -> None: ...

    @abc.abstractmethod
    async def clear_expired(self) -> None: ...


class MemoryCache(BaseTTLCache):
    def __init__(self) -> None:
        self._cache: dict[str, tuple[float, str]] = {}

    async def start(self) -> None: ...

    async def close(self) -> None: ...

    async def get(self, key: str) -> str | None:
        cached = self._cache.get(key)
        if cached is None:
            return None
        return cached[1]

    async def set(self, key: str, value: str, ttl: int) -> None:
        self._cache[key] = (time.time() + ttl, value)

    async def delete(self, key: str) -> None:
        self._cache.pop(key, None)

    async def clear_expired(self) -> None:
        now = time.time()

        for key, (expires_at, _) in list(self._cache.items()):
            if now >= expires_at:
                await self.delete(key)


class SQLiteCache(BaseTTLCache):
    def __init__(self, db_path: pathlib.Path | str = ".cache/enka_py.db") -> None:
        self._db_path = pathlib.Path(db_path)
        self._conn: aiosqlite.Connection | None = None

    @property
    def conn(self) -> aiosqlite.Connection:
        if self._conn is None:
            msg = f"Cache is not started, call `{self.__class__.__name__}.start` first"
            raise RuntimeError(msg)
        return self._conn

    async def start(self) -> None:
        os.makedirs(self._db_path.parent, exist_ok=True)
        self._conn = await aiosqlite.connect(self._db_path)
        await self.conn.execute(
            "CREATE TABLE IF NOT EXISTS cache (key TEXT PRIMARY KEY, value TEXT, expires_at REAL)"
        )
        await self.conn.commit()

    async def close(self) -> None:
        if self._conn is None:
            return
        await self._conn.close()

    async def get(self, key: str) -> str | None:
        async with self.conn.execute("SELECT value FROM cache WHERE key = ?", (key,)) as cursor:
            row = await cursor.fetchone()
            if row is None:
                return None
            return row[0]

    async def set(self, key: str, value: str, ttl: int) -> None:
        await self.conn.execute(
            "INSERT INTO cache (key, value, expires_at) VALUES (?, ?, ?) ON CONFLICT(key) DO UPDATE SET value = ?, expires_at = ?",
            (key, value, time.time() + ttl, value, time.time() + ttl),
        )
        await self.conn.commit()

    async def delete(self, key: str) -> None:
        await self.conn.execute("DELETE FROM cache WHERE key = ?", (key,))
        await self.conn.commit()

    async def clear_expired(self) -> None:
        await self.conn.execute("DELETE FROM cache WHERE expires_at < ?", (time.time(),))
        await self.conn.commit()



from __future__ import annotations

import contextlib
from typing import Any

import aiofiles
import orjson


class AssetData:
    """Base class for asset data."""

    def __init__(self) -> None:
        self._data: dict[str, Any] | None = None

    def __getitem__(self, key: str) -> Any:
        if self._data is None:
            msg = f"{self.__class__.__name__} not loaded"
            raise RuntimeError(msg)

        text = self._data.get(str(key))
        if text is None:
            msg = f"Cannot find text for key {key!r} in `{self.__class__.__name__}._data`, consider calling `update_assets` to update the assets"
            raise KeyError(msg)

        return text

    def __iter__(self) -> Any:
        if self._data is None:
            msg = f"{self.__class__.__name__} not loaded"
            raise RuntimeError(msg)

        return iter(self._data)

    def values(self) -> Any:
        """Get the values of the data."""
        if self._data is None:
            msg = f"{self.__class__.__name__} not loaded"
            raise RuntimeError(msg)

        return self._data.values()

    def items(self) -> Any:
        """Get the items of the data."""
        if self._data is None:
            msg = f"{self.__class__.__name__} not loaded"
            raise RuntimeError(msg)

        return self._data.items()

    async def _open_json(self, path: str) -> dict[str, Any] | None:
        with contextlib.suppress(FileNotFoundError):
            async with aiofiles.open(path, encoding="utf-8") as f:
                return orjson.loads(await f.read())
        return None

    def get(self, key: str, default: Any = None) -> str | Any:
        """Get a text by key.

        Args:
            key (str): The key to get the text for.
            default (Any): The default value to return if the key is not found.

        Returns:
            str | Any: The text or the default value.
        """
        if self._data is None:
            msg = f"{self.__class__.__name__} not loaded"
            raise RuntimeError(msg)

        text = self._data.get(str(key))
        if text is None:
            return default

        return text



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data import AssetData
from .file_paths import (
    CHARACTER_DATA_PATH,
    CONSTS_DATA_PATH,
    NAMECARD_DATA_PATH,
    PFPS_DATA_PATH,
    TALENTS_DATA_PATH,
    TEXT_MAP_PATH,
)

if TYPE_CHECKING:
    from ...enums.gi import Language

__all__ = ("AssetManager",)


class AssetManager:
    """Genshin Impact asset manager."""

    def __init__(self, lang: Language) -> None:
        self._lang = lang
        self.text_map = TextMap(lang)
        self.character_data = CharacterData()
        self.namecard_data = NamecardData()
        self.consts_data = ConstsData()
        self.talents_data = TalentsData()
        self.pfps_data = PfpsData()

    async def load(self) -> bool:
        """Load all assets.

        Returns:
            bool: Whether all assets were loaded successfully.
        """
        return (
            await self.text_map.load()
            and await self.character_data.load()
            and await self.namecard_data.load()
            and await self.consts_data.load()
            and await self.talents_data.load()
            and await self.pfps_data.load()
        )


class TextMap(AssetData):
    def __init__(self, lang: Language) -> None:
        super().__init__()
        self._lang = lang

    async def load(self) -> bool:
        self._data = await self._open_json(TEXT_MAP_PATH.format(lang=self._lang.value))
        return self._data is not None


class CharacterData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(CHARACTER_DATA_PATH)
        return self._data is not None


class NamecardData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(NAMECARD_DATA_PATH)
        return self._data is not None


class ConstsData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(CONSTS_DATA_PATH)
        return self._data is not None


class TalentsData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(TALENTS_DATA_PATH)
        return self._data is not None


class PfpsData(AssetData):
    async def load(self) -> bool:
        self._data = await self._open_json(PFPS_DATA_PATH)
        return self._data is not None



from __future__ import annotations

import math


def round_down(number: int | float, decimal_places: int) -> int | float:
    """Round down a number to a specified number of decimal places.

    Args:
        number: The number to round down.
        decimal_places: The number of decimal places to round down to.

    Returns:
        int | float: The rounded down number.
    """
    factor = 10**decimal_places
    return math.floor(number * factor) / factor


def update_stats(original: dict[str, float], new: dict[str, float]) -> None:
    """Update a dictionary of stats with a new dictionary of stats.

    Args:
        original: The original dictionary of stats.
        new: The new dictionary of stats.
    """
    for key, value in new.items():
        if key in original:
            original[key] += value
        else:
            original[key] = value



from __future__ import annotations

from . import errors as errors
from . import gi as gi
from . import hsr as hsr
from . import utils as utils
from .clients import GenshinClient as GenshinClient
from .clients import HSRClient as HSRClient
from .clients import cache as cache
from .enums.enum import Game as Game
from .models.enka import *



from __future__ import annotations

__all__ = (
    "AlgoScrewedUpMassivelyError",
    "AssetUpdateError",
    "EnkaAPIError",
    "EnkaPyError",
    "GameMaintenanceError",
    "GeneralServerError",
    "InvalidItemTypeError",
    "PlayerDoesNotExistError",
    "RateLimitedError",
    "WrongUIDFormatError",
    "raise_for_retcode",
)


class EnkaAPIError(Exception):
    """Base exception class for Enka API."""

    def __str__(self) -> str:
        return "Unknown error"


class WrongUIDFormatError(EnkaAPIError):
    """Raised when the UID format is incorrect."""

    def __str__(self) -> str:
        return "UID must be a string of 9 digits"


class PlayerDoesNotExistError(EnkaAPIError):
    """Raised when the player does not exist."""

    def __str__(self) -> str:
        return "Player does not exist"


class GameMaintenanceError(EnkaAPIError):
    """Raised when the game is under maintenance."""

    def __str__(self) -> str:
        return "Game is under maintenance"


class RateLimitedError(EnkaAPIError):
    """Raised when the API rate limit is exceeded."""

    def __str__(self) -> str:
        return "Rate limited"


class GeneralServerError(EnkaAPIError):
    """Raised when a general server error occurs."""

    def __str__(self) -> str:
        return "General server error"


class AlgoScrewedUpMassivelyError(EnkaAPIError):
    """Raised when something goes wrong on the server side."""

    def __str__(self) -> str:
        return "Algo screwed up massively"


class GatewayTimeoutError(EnkaAPIError):
    """Raised when a gateway timeout occurs."""

    def __str__(self) -> str:
        return "Gateway timeout"


def raise_for_retcode(retcode: int) -> None:
    """Raises an exception based on the retcode."""
    match retcode:
        case 400:
            raise WrongUIDFormatError
        case 404:
            raise PlayerDoesNotExistError
        case 424:
            raise GameMaintenanceError
        case 429:
            raise RateLimitedError
        case 500:
            raise GeneralServerError
        case 503:
            raise AlgoScrewedUpMassivelyError
        case 504:
            raise GatewayTimeoutError
        case _:
            raise EnkaAPIError


class EnkaPyError(Exception):
    """Base exception class for enka.py."""

    def __str__(self) -> str:
        return "enka.py error"


class InvalidItemTypeError(EnkaPyError):
    """Raised when the item type is invalid."""

    def __str__(self) -> str:
        return "Invalid item type"


class AssetUpdateError(EnkaPyError):
    """Raised when the assets fail to update."""

    def __init__(self, status: int, url: str) -> None:
        self.status = status
        self.url = url

    def __str__(self) -> str:
        return f"Failed to update assets, status code: {self.status}, url: {self.url}"



from __future__ import annotations

ASSET_PATH = ".enka_py/assets/hsr"

TEXT_MAP_PATH = f"{ASSET_PATH}/text_map.json"
CHARACTER_DATA_PATH = f"{ASSET_PATH}/characters.json"
LIGHT_CONE_DATA_PATH = f"{ASSET_PATH}/light_cone.json"
RELIC_DATA_PATH = f"{ASSET_PATH}/relic.json"
SKILL_TREE_DATA_PATH = f"{ASSET_PATH}/skill_tree.json"
META_DATA_PATH = f"{ASSET_PATH}/promotions.json"
AVATAR_DATA_PATH = f"{ASSET_PATH}/avatars.json"
PROPERTY_CONFIG_PATH = f"{ASSET_PATH}/property_config.json"
EIDOLON_DATA_PATH = f"{ASSET_PATH}/eidolons.json"

ENKA_API_DOCS = "https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/hsr"
ENKA_PY_ASSETS = "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/hsr"

SOURCE_TO_PATH = {
    f"{ENKA_API_DOCS}/hsr.json": TEXT_MAP_PATH,
    f"{ENKA_API_DOCS}/honker_characters.json": CHARACTER_DATA_PATH,
    f"{ENKA_API_DOCS}/honker_weps.json": LIGHT_CONE_DATA_PATH,
    f"{ENKA_API_DOCS}/honker_relics.json": RELIC_DATA_PATH,
    f"{ENKA_API_DOCS}/honker_meta.json": META_DATA_PATH,
    f"{ENKA_API_DOCS}/honker_avatars.json": AVATAR_DATA_PATH,
    f"{ENKA_API_DOCS}/honker_ranks.json": EIDOLON_DATA_PATH,
    f"{ENKA_PY_ASSETS}/skill_tree.json": SKILL_TREE_DATA_PATH,
    f"{ENKA_PY_ASSETS}/property_config.json": PROPERTY_CONFIG_PATH,
}



from __future__ import annotations

import asyncio

import enka

FIGHT_PROPS_TO_SHOW = (
    enka.gi.FightPropType.FIGHT_PROP_MAX_HP,
    enka.gi.FightPropType.FIGHT_PROP_CUR_ATTACK,
    enka.gi.FightPropType.FIGHT_PROP_CUR_DEFENSE,
    enka.gi.FightPropType.FIGHT_PROP_ELEMENT_MASTERY,
    enka.gi.FightPropType.FIGHT_PROP_CRITICAL,
    enka.gi.FightPropType.FIGHT_PROP_CRITICAL_HURT,
    enka.gi.FightPropType.FIGHT_PROP_CHARGE_EFFICIENCY,
)


async def main() -> None:  # noqa: C901, D103
    async with enka.GenshinClient(enka.gi.Language.TRADITIONAL_CHINESE) as api:
        # Update assets
        await api.update_assets()

        try:
            response = await api.fetch_showcase(831335713)
        except enka.errors.PlayerDoesNotExistError:
            return print("Player does not exist.")
        except enka.errors.GameMaintenanceError:
            return print("Game is in maintenance.")

        print("Name:", response.player.nickname)
        print("Level:", response.player.level)
        print("Achievements:", response.player.achievements)
        print("Namecard:", response.player.namecard.full)
        print("Profile picture side icon:", response.player.profile_picture_icon.side)

        for character in response.characters:
            print("\n===============================\n")
            print(
                f"Lv. {character.level}/{character.max_level} {character.name} (C{character.constellations_unlocked})"
            )
            print(f"Rarity: {character.rarity} ★")
            print("Element:", character.element.name.title())
            print("Side icon:", character.icon.side)
            print(f"Talent levels: {'/'.join(str(talent.level) for talent in character.talents)}")
            if character.namecard is not None:
                print("Namecard:", character.namecard.full)

            if character.costume is not None:
                print("Costume side icon:", character.costume.icon.side)

            weapon = character.weapon
            print("\nWeapon:")
            print(f"Lv. {weapon.level}/{weapon.max_level} {weapon.name} (R{weapon.refinement})")
            print(f"Rarity: {weapon.rarity} ★")
            for stat in weapon.stats:
                print(stat.name, stat.formatted_value)

            print("\nStats:")
            for stat_type, stat in character.stats.items():
                if stat_type in FIGHT_PROPS_TO_SHOW:
                    print(stat.name, stat.formatted_value)
            dmg_bonus = character.highest_dmg_bonus_stat
            print(dmg_bonus.name, dmg_bonus.formatted_value)

            print("\nArtifacts:")
            for artifact in character.artifacts:
                main_stat = artifact.main_stat
                print(
                    f"Lv. {artifact.level} {artifact.name}: {main_stat.name} {main_stat.formatted_value}"
                )
                for substat in artifact.sub_stats:
                    print(f"- {substat.name} {substat.formatted_value}")
                print("")


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

__all__ = ("Icon", "Namecard")


from pydantic import BaseModel, computed_field


class Icon(BaseModel):
    side_icon_ui_path: str
    is_costume: bool = False

    @computed_field
    @property
    def icon_ui_path(self) -> str:
        """The icon UI path of the character.

        e.g. UI_AvatarIcon_Ambor
        """
        return self.side_icon_ui_path.replace("Side_", "")

    @computed_field
    @property
    def side(self) -> str:
        """The side icon of the character.

        e.g. https://enka.network/ui/UI_AvatarIcon_Side_Ambor.png
        """
        return f"https://enka.network/ui/{self.side_icon_ui_path}.png"

    @computed_field
    @property
    def circle(self) -> str:
        """The circle (round) icon of the character.

        e.g. https://enka.network/ui/UI_AvatarIcon_Ambor_Circle.png
        """
        return self.side.replace("Side_", "").replace(".png", "_Circle.png")

    @computed_field
    @property
    def gacha(self) -> str:
        """The gacha art of the character.

        e.g. https://enka.network/ui/UI_Gacha_AvatarImg_Ambor.png
        """
        return self.side.replace(
            "AvatarIcon_Side", "Costume" if self.is_costume else "Gacha_AvatarImg"
        )

    @computed_field
    @property
    def front(self) -> str:
        """The front icon of the character.

        e.g. https://enka.network/ui/UI_AvatarIcon_Ambor.png
        """
        return self.side.replace("Side_", "")


class Namecard(BaseModel):
    ui_path: str

    @computed_field
    @property
    def icon(self) -> str:
        """The namecard's icon.

        e.g. https://enka.network/ui/UI_NameCardIcon_0.png
        """
        return f"https://enka.network/ui/{self.ui_path.replace('NameCardPic', 'NameCardIcon').replace('_P', '')}.png"

    @computed_field
    @property
    def full(self) -> str:
        """The full namecard.

        e.g. https://enka.network/ui/UI_NameCardPic_0_P.png
        """
        return f"https://enka.network/ui/{self.ui_path}.png"



from __future__ import annotations

TEXT_MAP_PATH = ".enka_py/assets/text_map_{lang}.json"
CHARACTER_DATA_PATH = ".enka_py/assets/characters.json"
NAMECARD_DATA_PATH = ".enka_py/assets/namecards.json"
CONSTS_DATA_PATH = ".enka_py/assets/consts.json"
TALENTS_DATA_PATH = ".enka_py/assets/talents.json"
PFPS_DATA_PATH = ".enka_py/assets/pfps.json"

SOURCE_TO_PATH = {
    "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/text_map_{lang}.json": TEXT_MAP_PATH,
    "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/characters.json": CHARACTER_DATA_PATH,
    "https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/namecards.json": NAMECARD_DATA_PATH,
    "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/consts.json": CONSTS_DATA_PATH,
    "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/talents.json": TALENTS_DATA_PATH,
    "https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/pfps.json": PFPS_DATA_PATH,
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import aiohttp
import orjson
from loguru import logger

from ..errors import raise_for_retcode

if TYPE_CHECKING:
    from ..enums.enum import Game
    from .cache import BaseTTLCache


class BaseClient:
    """Base client with requesting capabilities."""

    def __init__(
        self,
        game: Game,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
    ) -> None:
        self.game = game

        self._headers = headers or {"User-Agent": "enka-py"}
        self._session: aiohttp.ClientSession | None = None
        self._cache = cache

    async def __aenter__(self) -> BaseClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def start(self) -> None:
        """Start the client."""
        self._session = aiohttp.ClientSession(headers=self._headers)
        if self._cache is not None:
            await self._cache.start()

    async def close(self) -> None:
        """Close the client."""
        if self._session is None:
            msg = f"Client is not started, call `{self.__class__.__name__}.start` first"
            raise RuntimeError(msg)

        await self._session.close()
        if self._cache is not None:
            await self._cache.close()

    async def _request(self, url: str) -> dict[str, Any]:
        if self._session is None:
            msg = f"Client is not started, call `{self.__class__.__name__}.start` first"
            raise RuntimeError(msg)

        if self._cache is not None:
            await self._cache.clear_expired()
            cached = await self._cache.get(url)
            if cached is not None:
                return orjson.loads(cached.encode())

        logger.debug(f"Requesting {url}")

        async with self._session.get(url) as resp:
            if resp.status != 200:
                raise_for_retcode(resp.status)

            data: dict[str, Any] = await resp.json()
            if self._cache is not None:
                await self._cache.set(url, orjson.dumps(data).decode(), ttl=60)
            return data



from __future__ import annotations

from typing import Literal

from ..enums.hsr import Element

__all__ = ("ASCENSION_TO_MAX_LEVEL", "DEFAULT_STATS", "DMG_BONUS_PROPS", "PERCENT_STAT_TYPES")

PERCENT_STAT_TYPES = {
    "CriticalChance",
    "CriticalDamage",
    "BreakDamageAddedRatio",
    "BreakDamageAddedRatioBase",
    "HealRatio",
    "SPRatio",
    "StatusProbability",
    "StatusResistance",
    "CriticalChanceBase",
    "CriticalDamageBase",
    "HealRatioBase",
    "SPRatioBase",
    "StatusProbabilityBase",
    "StatusResistanceBase",
    "PhysicalAddedRatio",
    "PhysicalResistance",
    "FireAddedRatio",
    "FireResistance",
    "IceAddedRatio",
    "IceResistance",
    "ThunderAddedRatio",
    "ThunderResistance",
    "WindAddedRatio",
    "WindResistance",
    "QuantumAddedRatio",
    "QuantumResistance",
    "ImaginaryAddedRatio",
    "ImaginaryResistance",
    "HPAddedRatio",
    "AttackAddedRatio",
    "DefenceAddedRatio",
    "HealTakenRatio",
}

ASCENSION_TO_MAX_LEVEL: dict[Literal[0, 1, 2, 3, 4, 5, 6], Literal[20, 30, 40, 50, 60, 70, 80]] = {
    0: 20,
    1: 30,
    2: 40,
    3: 50,
    4: 60,
    5: 70,
    6: 80,
}

DMG_BONUS_PROPS = {
    Element.PHYSICAL: "PhysicalAddedRatio",
    Element.FIRE: "FireAddedRatio",
    Element.LIGHTNING: "ThunderAddedRatio",
    Element.ICE: "IceAddedRatio",
    Element.QUANTUM: "QuantumAddedRatio",
    Element.WIND: "WindAddedRatio",
    Element.IMAGINARY: "ImaginaryAddedRatio",
}

DEFAULT_STATS: dict[str, float] = {
    "BaseHP": 0,
    "HPAddedRatio": 0,
    "HPDelta": 0,
    "HPConvert": 0,
    "BaseAttack": 0,
    "AttackAddedRatio": 0,
    "AttackDelta": 0,
    "AttackConvert": 0,
    "BaseDefence": 0,
    "DefenceAddedRatio": 0,
    "DefenceDelta": 0,
    "DefenceConvert": 0,
    "BaseSpeed": 0,
    "SpeedAddedRatio": 0,
    "SpeedDelta": 0,
    "SpeedConvert": 0,
    "CriticalChance": 0,
    "CriticalChanceBase": 0,
    "CriticalDamage": 0,
    "CriticalDamageBase": 0,
    "SPRatio": 0,
    "SPRatioBase": 0,
    "SPRatioConvert": 0,
    "StatusProbability": 0,
    "StatusProbabilityBase": 0,
    "StatusProbabilityConvert": 0,
    "StatusResistance": 0,
    "StatusResistanceBase": 0,
    "StatusResistanceConvert": 0,
    "HealRatioBase": 0,
    "HealRatioConvert": 0,
    "HealTakenRatio": 0,
    "ShieldAddedRatio": 0,
    "ShieldTakenRatio": 0,
    "AggroBase": 0,
    "AggroAddedRatio": 0,
    "AggroDelta": 0,
    "BreakDamageAddedRatio": 0,
    "BreakDamageAddedRatioBase": 0,
    "AllDamageTypeResistance": 0,
    "PhysicalResistanceDelta": 0,
    "FireResistanceDelta": 0,
    "IceResistanceDelta": 0,
    "ThunderResistanceDelta": 0,
    "QuantumResistanceDelta": 0,
    "ImaginaryResistanceDelta": 0,
    "WindResistanceDelta": 0,
    "PhysicalPenetrate": 0,
    "FirePenetrate": 0,
    "IcePenetrate": 0,
    "ThunderPenetrate": 0,
    "QuantumPenetrate": 0,
    "ImaginaryPenetrate": 0,
    "WindPenetrate": 0,
    "AllDamageTypeTakenRatio": 0,
    "PhysicalTakenRatio": 0,
    "FireTakenRatio": 0,
    "IceTakenRatio": 0,
    "ThunderTakenRatio": 0,
    "QuantumTakenRatio": 0,
    "ImaginaryTakenRatio": 0,
    "WindTakenRatio": 0,
    "AllDamageTypeAddedRatio": 0,
    "DOTDamageAddedRatio": 0,
    "PhysicalAddedRatio": 0,
    "FireAddedRatio": 0,
    "IceAddedRatio": 0,
    "ThunderAddedRatio": 0,
    "QuantumAddedRatio": 0,
    "ImaginaryAddedRatio": 0,
    "WindAddedRatio": 0,
    "StanceBreakAddedRatio": 0,
    "AllDamageReduce": 0,
    "FatigueRatio": 0,
    "MinimumFatigueRatio": 0,
}



from __future__ import annotations

import sys

if sys.version_info < (3, 11):
    from enum import Enum as StrEnum
else:
    from enum import StrEnum

__all__ = ("Game",)


class Game(StrEnum):
    """Game."""

    GI = "gi"
    HSR = "hsr"



from __future__ import annotations

from enka import HSRClient
from enka.hsr import Language


async def test_fetch_showcase() -> None:
    async with HSRClient() as api:
        await api.fetch_showcase("809162009")


async def test_trailblazer_showcase() -> None:
    async with HSRClient() as api:
        await api.fetch_showcase("123456786")


async def test_update_assets() -> None:
    async with HSRClient() as api:
        await api.update_assets()


async def test_langs() -> None:
    for lang in Language:
        async with HSRClient(lang) as api:
            await api.fetch_showcase("809162009")


async def test_low_level_acc() -> None:
    async with HSRClient() as api:
        await api.fetch_showcase("829702635")


async def test_owner_and_builds() -> None:
    async with HSRClient() as api:
        showcase = await api.fetch_showcase("809162009")
        assert showcase.owner is not None
        await api.fetch_builds(showcase.owner)


async def test_raw_and_parse() -> None:
    async with HSRClient() as api:
        raw = await api.fetch_showcase("809162009", raw=True)
        api.parse_showcase(raw)


async def test_empty_substat_list() -> None:
    async with HSRClient() as api:
        await api.fetch_showcase("800724088")



from __future__ import annotations

import sys
from enum import IntEnum

if sys.version_info < (3, 11):
    from enum import Enum as StrEnum
else:
    from enum import StrEnum

__all__ = ("Element", "EquipmentType", "FightPropType", "ItemType", "Language", "StatType")


class Language(StrEnum):
    """Genshin Impact API language codes."""

    ENGLISH = "en"
    RUSSIAN = "ru"
    VIETNAMESE = "vi"
    THAI = "th"
    PORTUGUESE = "pt"
    KOREAN = "ko"
    JAPANESE = "ja"
    INDONESIAN = "id"
    FRENCH = "fr"
    SPANISH = "es"
    GERMAN = "de"
    TRADITIONAL_CHINESE = "zh-tw"
    SIMPLIFIED_CHINESE = "zh-cn"
    ITALIAN = "it"
    TURKISH = "tr"


class EquipmentType(StrEnum):
    """GI artifact types."""

    FLOWER = "EQUIP_BRACER"
    FEATHER = "EQUIP_NECKLACE"
    SANDS = "EQUIP_SHOES"
    GOBLET = "EQUIP_RING"
    CIRCLET = "EQUIP_DRESS"


class ItemType(StrEnum):
    """GI item types."""

    WEAPON = "ITEM_WEAPON"
    ARTIFACT = "ITEM_RELIQUARY"


class StatType(StrEnum):
    """GI stat types."""

    FIGHT_PROP_BASE_ATTACK = "FIGHT_PROP_BASE_ATTACK"
    FIGHT_PROP_HP = "FIGHT_PROP_HP"
    FIGHT_PROP_ATTACK = "FIGHT_PROP_ATTACK"
    FIGHT_PROP_DEFENSE = "FIGHT_PROP_DEFENSE"
    FIGHT_PROP_HP_PERCENT = "FIGHT_PROP_HP_PERCENT"
    FIGHT_PROP_ATTACK_PERCENT = "FIGHT_PROP_ATTACK_PERCENT"
    FIGHT_PROP_DEFENSE_PERCENT = "FIGHT_PROP_DEFENSE_PERCENT"
    FIGHT_PROP_CRITICAL = "FIGHT_PROP_CRITICAL"
    FIGHT_PROP_CRITICAL_HURT = "FIGHT_PROP_CRITICAL_HURT"
    FIGHT_PROP_CHARGE_EFFICIENCY = "FIGHT_PROP_CHARGE_EFFICIENCY"
    FIGHT_PROP_HEAL_ADD = "FIGHT_PROP_HEAL_ADD"
    FIGHT_PROP_ELEMENT_MASTERY = "FIGHT_PROP_ELEMENT_MASTERY"
    FIGHT_PROP_PHYSICAL_ADD_HURT = "FIGHT_PROP_PHYSICAL_ADD_HURT"
    FIGHT_PROP_FIRE_ADD_HURT = "FIGHT_PROP_FIRE_ADD_HURT"
    FIGHT_PROP_ELEC_ADD_HURT = "FIGHT_PROP_ELEC_ADD_HURT"
    FIGHT_PROP_WATER_ADD_HURT = "FIGHT_PROP_WATER_ADD_HURT"
    FIGHT_PROP_WIND_ADD_HURT = "FIGHT_PROP_WIND_ADD_HURT"
    FIGHT_PROP_ICE_ADD_HURT = "FIGHT_PROP_ICE_ADD_HURT"
    FIGHT_PROP_ROCK_ADD_HURT = "FIGHT_PROP_ROCK_ADD_HURT"
    FIGHT_PROP_GRASS_ADD_HURT = "FIGHT_PROP_GRASS_ADD_HURT"


class FightPropType(IntEnum):
    """GI fight property types."""

    FIGHT_PROP_NONE = 0
    FIGHT_PROP_BASE_HP = 1
    FIGHT_PROP_HP = 2
    FIGHT_PROP_HP_PERCENT = 3
    FIGHT_PROP_BASE_ATTACK = 4
    FIGHT_PROP_ATTACK = 5
    FIGHT_PROP_ATTACK_PERCENT = 6
    FIGHT_PROP_BASE_DEFENSE = 7
    FIGHT_PROP_DEFENSE = 8
    FIGHT_PROP_DEFENSE_PERCENT = 9
    FIGHT_PROP_BASE_SPEED = 10
    FIGHT_PROP_SPEED_PERCENT = 11
    FIGHT_PROP_HP_MP_PERCENT = 12
    FIGHT_PROP_ATTACK_MP_PERCENT = 13
    FIGHT_PROP_CRITICAL = 20
    FIGHT_PROP_ANTI_CRITICAL = 21
    FIGHT_PROP_CRITICAL_HURT = 22
    FIGHT_PROP_CHARGE_EFFICIENCY = 23
    FIGHT_PROP_ADD_HURT = 24
    FIGHT_PROP_SUB_HURT = 25
    FIGHT_PROP_HEAL_ADD = 26
    FIGHT_PROP_HEALED_ADD = 27
    FIGHT_PROP_ELEMENT_MASTERY = 28
    FIGHT_PROP_PHYSICAL_SUB_HURT = 29
    FIGHT_PROP_PHYSICAL_ADD_HURT = 30
    FIGHT_PROP_DEFENCE_IGNORE_RATIO = 31
    FIGHT_PROP_DEFENCE_IGNORE_DELTA = 32
    FIGHT_PROP_FIRE_ADD_HURT = 40
    FIGHT_PROP_ELEC_ADD_HURT = 41
    FIGHT_PROP_WATER_ADD_HURT = 42
    FIGHT_PROP_GRASS_ADD_HURT = 43
    FIGHT_PROP_WIND_ADD_HURT = 44
    FIGHT_PROP_ROCK_ADD_HURT = 45
    FIGHT_PROP_ICE_ADD_HURT = 46
    FIGHT_PROP_HIT_HEAD_ADD_HURT = 47
    FIGHT_PROP_FIRE_SUB_HURT = 50
    FIGHT_PROP_ELEC_SUB_HURT = 51
    FIGHT_PROP_WATER_SUB_HURT = 52
    FIGHT_PROP_GRASS_SUB_HURT = 53
    FIGHT_PROP_WIND_SUB_HURT = 54
    FIGHT_PROP_ROCK_SUB_HURT = 55
    FIGHT_PROP_ICE_SUB_HURT = 56
    FIGHT_PROP_EFFECT_HIT = 60
    FIGHT_PROP_EFFECT_RESIST = 61
    FIGHT_PROP_FREEZE_RESIST = 62
    FIGHT_PROP_DIZZY_RESIST = 64
    FIGHT_PROP_FREEZE_SHORTEN = 65
    FIGHT_PROP_DIZZY_SHORTEN = 67
    FIGHT_PROP_MAX_FIRE_ENERGY = 70
    FIGHT_PROP_MAX_ELEC_ENERGY = 71
    FIGHT_PROP_MAX_WATER_ENERGY = 72
    FIGHT_PROP_MAX_GRASS_ENERGY = 73
    FIGHT_PROP_MAX_WIND_ENERGY = 74
    FIGHT_PROP_MAX_ICE_ENERGY = 75
    FIGHT_PROP_MAX_ROCK_ENERGY = 76
    FIGHT_PROP_SKILL_CD_MINUS_RATIO = 80
    FIGHT_PROP_SHIELD_COST_MINUS_RATIO = 81
    FIGHT_PROP_CUR_FIRE_ENERGY = 1000
    FIGHT_PROP_CUR_ELEC_ENERGY = 1001
    FIGHT_PROP_CUR_WATER_ENERGY = 1002
    FIGHT_PROP_CUR_GRASS_ENERGY = 1003
    FIGHT_PROP_CUR_WIND_ENERGY = 1004
    FIGHT_PROP_CUR_ICE_ENERGY = 1005
    FIGHT_PROP_CUR_ROCK_ENERGY = 1006
    FIGHT_PROP_CUR_HP = 1010
    FIGHT_PROP_MAX_HP = 2000
    FIGHT_PROP_CUR_ATTACK = 2001
    FIGHT_PROP_CUR_DEFENSE = 2002
    FIGHT_PROP_CUR_SPEED = 2003
    FIGHT_PROP_CUR_HP_DEBTS = 2004
    FIGHT_PROP_CUR_HP_PAID_DEBTS = 2005
    FIGHT_PROP_NONEXTRA_ATTACK = 3000
    FIGHT_PROP_NONEXTRA_DEFENSE = 3001
    FIGHT_PROP_NONEXTRA_CRITICAL = 3002
    FIGHT_PROP_NONEXTRA_ANTI_CRITICAL = 3003
    FIGHT_PROP_NONEXTRA_CRITICAL_HURT = 3004
    FIGHT_PROP_NONEXTRA_CHARGE_EFFICIENCY = 3005
    FIGHT_PROP_NONEXTRA_ELEMENT_MASTERY = 3006
    FIGHT_PROP_NONEXTRA_PHYSICAL_SUB_HURT = 3007
    FIGHT_PROP_NONEXTRA_FIRE_ADD_HURT = 3008
    FIGHT_PROP_NONEXTRA_ELEC_ADD_HURT = 3009
    FIGHT_PROP_NONEXTRA_WATER_ADD_HURT = 3010
    FIGHT_PROP_NONEXTRA_GRASS_ADD_HURT = 3011
    FIGHT_PROP_NONEXTRA_WIND_ADD_HURT = 3012
    FIGHT_PROP_NONEXTRA_ROCK_ADD_HURT = 3013
    FIGHT_PROP_NONEXTRA_ICE_ADD_HURT = 3014
    FIGHT_PROP_NONEXTRA_FIRE_SUB_HURT = 3015
    FIGHT_PROP_NONEXTRA_ELEC_SUB_HURT = 3016
    FIGHT_PROP_NONEXTRA_WATER_SUB_HURT = 3017
    FIGHT_PROP_NONEXTRA_GRASS_SUB_HURT = 3018
    FIGHT_PROP_NONEXTRA_WIND_SUB_HURT = 3019
    FIGHT_PROP_NONEXTRA_ROCK_SUB_HURT = 3020
    FIGHT_PROP_NONEXTRA_ICE_SUB_HURT = 3021
    FIGHT_PROP_NONEXTRA_SKILL_CD_MINUS_RATIO = 3022
    FIGHT_PROP_NONEXTRA_SHIELD_COST_MINUS_RATIO = 3023
    FIGHT_PROP_NONEXTRA_PHYSICAL_ADD_HURT = 3024
    FIGHT_PROP_BASE_ELEM_REACT_CRITICAL = 3045
    FIGHT_PROP_BASE_ELEM_REACT_CRITICAL_HURT = 3046
    FIGHT_PROP_ELEM_REACT_CRITICAL = 3025
    FIGHT_PROP_ELEM_REACT_CRITICAL_HURT = 3026
    FIGHT_PROP_ELEM_REACT_EXPLODE_CRITICAL = 3027
    FIGHT_PROP_ELEM_REACT_EXPLODE_CRITICAL_HURT = 3028
    FIGHT_PROP_ELEM_REACT_SWIRL_CRITICAL = 3029
    FIGHT_PROP_ELEM_REACT_SWIRL_CRITICAL_HURT = 3030
    FIGHT_PROP_ELEM_REACT_ELECTRIC_CRITICAL = 3031
    FIGHT_PROP_ELEM_REACT_ELECTRIC_CRITICAL_HURT = 3032
    FIGHT_PROP_ELEM_REACT_SCONDUCT_CRITICAL = 3033
    FIGHT_PROP_ELEM_REACT_SCONDUCT_CRITICAL_HURT = 3034
    FIGHT_PROP_ELEM_REACT_BURN_CRITICAL = 3035
    FIGHT_PROP_ELEM_REACT_BURN_CRITICAL_HURT = 3036
    FIGHT_PROP_ELEM_REACT_FROZENBROKEN_CRITICAL = 3037
    FIGHT_PROP_ELEM_REACT_FROZENBROKEN_CRITICAL_HURT = 3038
    FIGHT_PROP_ELEM_REACT_OVERGROW_CRITICAL = 3039
    FIGHT_PROP_ELEM_REACT_OVERGROW_CRITICAL_HURT = 3040
    FIGHT_PROP_ELEM_REACT_OVERGROW_FIRE_CRITICAL = 3041
    FIGHT_PROP_ELEM_REACT_OVERGROW_FIRE_CRITICAL_HURT = 3042
    FIGHT_PROP_ELEM_REACT_OVERGROW_ELECTRIC_CRITICAL = 3043
    FIGHT_PROP_ELEM_REACT_OVERGROW_ELECTRIC_CRITICAL_HURT = 3044


class Element(StrEnum):
    """Genshin Impact elements."""

    ANEMO = "Wind"
    GEO = "Rock"
    ELECTRO = "Electric"
    DENDRO = "Grass"
    PYRO = "Fire"
    CRYO = "Ice"
    HYDRO = "Water"



from __future__ import annotations

from enka import GenshinClient
from enka.gi import Language


async def test_fetch_showcase() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("901211014")
        assert showcase.uid == "901211014"


async def test_empty_showcase() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("123456789")
        assert showcase.uid == "123456789"
        assert len(showcase.characters) == 0


async def test_traveler_showcase() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("600001919")
        assert showcase.uid == "600001919"


async def test_update_assets() -> None:
    async with GenshinClient() as api:
        await api.update_assets()


async def test_zero_achievement() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("831335714")
        assert showcase.player.achievements == 0


async def test_new_profile_picture_format() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("724824926")
        assert showcase.player.profile_picture_icon is not None


async def test_costume() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("738081787")
        assert showcase.player.showcase_characters[3].costume is not None


async def test_langs() -> None:
    for lang in Language:
        async with GenshinClient(lang) as api:
            await api.fetch_showcase("901211014")


async def test_owner_and_builds() -> None:
    async with GenshinClient() as api:
        showcase = await api.fetch_showcase("618285856")
        assert showcase.owner is not None
        await api.fetch_builds(showcase.owner)


async def test_raw_and_parse() -> None:
    async with GenshinClient() as api:
        raw = await api.fetch_showcase("901211014", raw=True)
        api.parse_showcase(raw)


async def test_invalid_player() -> None:
    async with GenshinClient() as api:
        await api.fetch_showcase("901211015")



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator

from .costume import Costume
from .icon import Icon, Namecard

__all__ = ("Player", "ShowcaseCharacter")


class ShowcaseCharacter(BaseModel):
    """
    Represents a showcase character.

    Attributes:
        id (int): The character's ID.
        level (int): The character's level.
        costume (Optional[Costume]): The character's costume, if any.
        costume_id (Optional[int]): The character's costume's ID, if any.
    """

    id: int = Field(alias="avatarId")
    level: int
    costume: Costume | None = None
    costume_id: int | None = Field(None, alias="costumeId")

    model_config = {"arbitrary_types_allowed": True}


class Player(BaseModel):
    """Represents a Genshin Impact player.

    Attributes:
        achievements (int): The number of completed achievements.
        level (int): The player's adventure level.
        namecard_id (int): The player's namecard's ID.
        namecard (Namecard): The player's namecard.
        nickname (str, optional): The player's nickname.
        signature (str, optional): The player's signature.
        abyss_floor (int): The player's Spiral Abyss floor.
        abyss_level (int): The player's Spiral Abyss level.
        world_level (int): The player's world level.
        profile_picture_id (int): The player's profile picture's ID.
        profile_picture_icon (Icon): The player's profile picture's icon.
        showcase_characters (List[ShowcaseCharacter]): The player's showcase characters.
    """

    achievements: int = Field(0, alias="finishAchievementNum")
    level: int
    namecard_id: int = Field(alias="nameCardId")
    namecard: Namecard = Namecard(ui_path="")
    nickname: str | None = None
    signature: str | None = ""
    abyss_floor: int = Field(0, alias="towerFloorIndex")
    abyss_level: int = Field(0, alias="towerLevelIndex")
    world_level: int = Field(0, alias="worldLevel")
    profile_picture_id: int = Field(alias="profilePicture")
    profile_picture_icon: Icon = Icon(side_icon_ui_path="")
    showcase_characters: list[ShowcaseCharacter] = Field([], alias="showAvatarInfoList")

    # New fields after v5.0
    max_friendship_character_count: int | None = Field(None, alias="fetterCount")
    abyss_stars: int | None = Field(None, alias="towerStarIndex")
    theater_stars: int | None = Field(None, alias="theaterStarIndex")

    model_config = {"arbitrary_types_allowed": True}

    @field_validator("profile_picture_id", mode="before")
    def _extract_avatar_id(cls, v: dict[str, int]) -> int:
        if not v:
            return 10000007

        avatar_id = v.get("avatarId", v.get("id"))
        if avatar_id is None:
            msg = f"Cannot find profile picture ID from {v}, maybe there is a new format?"
            raise KeyError(msg)
        return avatar_id



from __future__ import annotations

import asyncio

import enka


async def main() -> None:
    cache = enka.cache.SQLiteCache()
    async with enka.GenshinClient(cache=cache) as client:
        await client.fetch_showcase(901211014)  # Cached
        await client.fetch_showcase(901211014)  # From cache

    cache = enka.cache.MemoryCache()
    async with enka.GenshinClient(cache=cache) as client:
        await client.fetch_showcase(901211014)  # Cached
        await client.fetch_showcase(901211014)  # From cache


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_validator, model_validator

from ...constants.hsr import ASCENSION_TO_MAX_LEVEL, DMG_BONUS_PROPS, PERCENT_STAT_TYPES
from ...enums.hsr import Element, Path, RelicType, StatType
from ...utils import round_down
from .icon import CharacterIcon, LightConeIcon

__all__ = ("Character", "LightCone", "Relic", "RelicSubAffix", "Stat", "Trace")


class Trace(BaseModel):
    id: int = Field(alias="pointId")
    level: int
    boosted: bool = False
    """Whether the level of this trace is boosted by an activated eidolon's effect."""

    # Following fields are added in post-processing
    icon: str = ""
    max_level: int = 0
    anchor: str = ""
    type: int = 0


class Stat(BaseModel):
    type: StatType
    value: float

    # Following fields are added in post-processing
    name: str = ""
    icon: str = ""

    @computed_field
    @property
    def is_percentage(self) -> bool:
        """Whether the stat is a percentage stat."""
        return self.type.value in PERCENT_STAT_TYPES

    @computed_field
    @property
    def formatted_value(self) -> str:
        """Returns the formatted value of the stat."""
        if self.is_percentage:
            return f"{round_down(self.value * 100, 1)}%"
        else:
            if self.type in {StatType.SPD, StatType.SPEED_DELTA}:
                return str(round_down(self.value, 2))
            return str(int(round_down(self.value, 0)))


class LightCone(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: int = Field(alias="tid")
    level: int
    ascension: Literal[0, 1, 2, 3, 4, 5, 6] = Field(0, alias="promotion")
    superimpose: Literal[1, 2, 3, 4, 5] = Field(alias="rank")
    name: str  # Returned as text map hash in the API response
    stats: list[Stat] = Field(alias="props")

    # Following fields are added in post-processing
    icon: LightConeIcon = LightConeIcon(light_cone_id=0)
    rarity: Literal[3, 4, 5] = 3

    @computed_field
    @property
    def max_level(self) -> Literal[20, 30, 40, 50, 60, 70, 80]:
        """Light Cone's max level."""
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @field_validator("name", mode="before")
    def _stringify_name(cls, value: int) -> str:
        return str(value)

    @model_validator(mode="before")
    def _flatten_flat(cls, values: dict[str, Any]) -> dict[str, Any]:
        flat_ = values.pop("_flat")
        values.update(flat_)
        return values


class RelicSubAffix(BaseModel):
    id: int = Field(alias="affixId")
    cnt: int
    step: int | None = None


class Relic(BaseModel):
    id: int = Field(alias="tid")
    level: int = 0
    type: RelicType
    main_affix_id: int = Field(alias="mainAffixId")
    set_name: str = Field(alias="setName")  # Returned as text map hash in the API response
    set_id: int = Field(alias="setID")
    stats: list[Stat] = Field(alias="props")
    sub_affix_list: list[RelicSubAffix] = Field(alias="subAffixList", default_factory=list)

    # The following fields are added in post-processing
    icon: str = ""
    rarity: Literal[3, 4, 5] = 3

    @field_validator("set_name", mode="before")
    def _stringify_set_name(cls, value: int) -> str:
        return str(value)

    @model_validator(mode="before")
    def _flatten_flat(cls, values: dict[str, Any]) -> dict[str, Any]:
        flat_ = values.pop("_flat")
        values.update(flat_)
        return values

    @computed_field
    @property
    def main_stat(self) -> Stat:
        return self.stats[0]

    @computed_field
    @property
    def sub_stats(self) -> list[Stat]:
        return self.stats[1:]


class Eidolon(BaseModel):
    id: int
    icon: str
    unlocked: bool


class Character(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    level: int
    ascension: Literal[0, 1, 2, 3, 4, 5, 6] = Field(0, alias="promotion")
    id: int = Field(alias="avatarId")
    traces: list[Trace] = Field(alias="skillTreeList")
    light_cone: LightCone | None = Field(None, alias="equipment")
    relics: list[Relic] = Field(alias="relicList", default_factory=list)
    eidolons: list[Eidolon] = Field(default_factory=list)
    eidolons_unlocked: int = Field(0, alias="rank")
    is_assist: bool = Field(False, alias="_assist")

    # Following fields are added in post-processing
    icon: CharacterIcon = CharacterIcon(character_id=0)
    name: str = ""
    rarity: Literal[4, 5] = 4
    element: Element = Element.FIRE
    path: Path = Path.ABUNDANCE
    stats: dict[StatType, Stat] = Field(default_factory=dict)

    @computed_field
    @property
    def max_level(self) -> Literal[20, 30, 40, 50, 60, 70, 80]:
        """Character's max level."""
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @computed_field
    @property
    def highest_dmg_bonus_stat(self) -> Stat:
        """Character's highest damage bonus stat."""
        return max(
            (stat for stat in self.stats.values() if stat.type in DMG_BONUS_PROPS.values()),
            key=lambda stat: stat.value,
            default=next(
                stat
                for stat in self.stats.values()
                if stat.type.value == DMG_BONUS_PROPS[self.element]
            ),
        )



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, model_validator

from ..enka.owner import Owner
from .character import Character
from .player import Player

__all__ = ("ShowcaseResponse",)


class ShowcaseResponse(BaseModel):
    """
    Represents a Genshin Impact showcase response.

    Attributes:
        characters (List[Character]): The characters in the showcase.
        player (Player): The player of the showcase.
        ttl (int): The time to live of the response.
        uid (str): The UID of the showcase.
        owner (Owner, optional): The owner of the showcase's account.
    """

    characters: list[Character] = Field(alias="avatarInfoList")
    player: Player = Field(alias="playerInfo")
    ttl: int
    uid: str
    owner: Owner | None = None

    @model_validator(mode="before")
    def _handle_no_showcase(cls, v: dict[str, Any]) -> dict[str, Any]:
        if "avatarInfoList" not in v or v["avatarInfoList"] is None:
            v["avatarInfoList"] = []
        return v
